/* DSP LABORATORY 
 * University of Technology Graz, 2018
 * Josef Kulmer
 * 
 * UNIT 1, part 3: Quantization using several depths in bits
 * 
 * 
 */
#include <iostream>
#include <math.h>
#include "portaudio.h"
#include "blockprocessing.h"
#include <vector>
#include<cmath>

struct UserDataExperiment{
    short MASK=0x7FFF;
    /*  Change this mask for different quantization levels 			    		*/
    /*  1-bit quantizer   0x7FFF 											*/
    /*  2-bit quantizer   0x3FFF 											*/
    /*  3-bit quantizer   0x1FFF  	 										*/
    /*  4-bit quantizer   0x0FFF  	 										*/
    std::vector<float> qinF; // stores the quantized input
    int checkCallback;
};

// the callback function, evaluated at each interrupt
static int paCallback(const void *input, void *output,
    unsigned long framelength,
    const PaStreamCallbackTimeInfo* timeInfo,
    PaStreamCallbackFlags statusFlags,
    void *userdata)
{
    float *in = (float*)input;
    float *out = (float*)output;
    UserDataExperiment *ud = (UserDataExperiment*)userdata;

    short qinS;
    float meanErr=0, meanSig=0, enErr=0, enSig=0;
    
    // Quantize input and forward to output
    for(int i=0;i<framelength;i++){
        qinS = static_cast<short>(Q15 * in[i]);
        if ( qinS>=0 )
            qinS=qinS | ud->MASK;
	else
            qinS=qinS & (~ud->MASK);
        ud->qinF[i] = (static_cast<float>(qinS)/Q15);
        *out++ = ud->qinF[i];
    }
    
    // Calculate error measure
    meanErr=0;
    meanSig=0;
    enErr=0;
    enSig=0;
    
    // get mean value
    for(int i=0;i<framelength;i++){
        meanSig += in[i];
        meanErr += in[i]-ud->qinF[i];
    }
    meanSig /= framelength;
    meanErr /= framelength;
    
    // subtract mean and calculate power
    for(int i=0;i<framelength;i++){
        enSig += pow(in[i]-meanSig,2);
        enErr += pow(in[i]-ud->qinF[i]-meanErr,2);
    }
    enSig /= framelength;
    enErr /= framelength;
    
    printf("Signal power: %f / Quantization power: %f \n",enSig,enErr);
    
    ud->checkCallback ++; // report that callback function is working

    return 0;
}

int main(int argc, char *argv[]) {   
    PaStream *stream;
    PaError err;
    int checkCallback_local = 0; // check whether callback function is still active
    int iTimecounter = 0;    
    UserDataExperiment ud; // create objects
    ud.qinF.resize(FRAMELENGTH);
    ud.checkCallback = 0; // a watchdog if callback is still working (necessary for RPi) 
    
    // the following loop stops and restarts the portaudio stream if it is broken
    do{
        err = Pa_Initialize();
        std::cout << "Initialize portaudio: " << Pa_GetErrorText( err ) << std::endl;

        err = Pa_OpenDefaultStream(&stream, 
            NINPUTCHANNELS, 
            NOUTPUTCHANNELS, 
            paFloat32, 
            FS, 
            FRAMELENGTH, 
            paCallback, 
            &ud);
        std::cout << "Open audio stream: " << Pa_GetErrorText( err ) << std::endl;

        err = Pa_StartStream(stream);
        std::cout << "Start audio stream: " << Pa_GetErrorText( err ) << std::endl;

        Pa_Sleep(TIMETOSLEEP); // wait until Stream is open
        
        // wait in this loop until DURATION is reached or callback function needs to be restarted
        while(iTimecounter < DURATION-1 && ud.checkCallback != checkCallback_local){ // there was a callback within last second
            checkCallback_local = ud.checkCallback;
            Pa_Sleep(TIMETOSLEEP); // wait a second
            iTimecounter++;
        }
        
        if(iTimecounter != DURATION-1 && ud.checkCallback != checkCallback_local){
            std::cout << "There is a problem with portaudio -> restart" << std::endl;
        }

        err = Pa_StopStream(stream);
        std::cout << "Stop audio stream: " << Pa_GetErrorText( err ) << std::endl;

        err = Pa_CloseStream(stream);
        std::cout << "Close audio stream: " << Pa_GetErrorText( err ) << std::endl;
    }while(iTimecounter < DURATION-1);

    Pa_Terminate();
    return 0;    
}
